from __future__ import annotations

# Public surface for skill docs as packaged resources.
#
# Consumers (including `jaunt skill export`) should load these via
# `importlib.resources.files("jaunt.skill")` rather than assuming filesystem paths.

__all__ = []
