from __future__ import annotations

from collections.abc import Callable
from typing import cast

import pytest

# Example test specs.
# These tests express behavior and edge cases; the implementation is generated by Jaunt.


def test_slugify_basic() -> None:
    # Replace this import path with your project's module after you run `jaunt build`.
    #
    # from myproj.__generated__.text import slugify
    slugify = cast(Callable[[str], str], None)

    assert slugify("Hello World") == "hello-world"


def test_slugify_rejects_empty() -> None:
    slugify = cast(Callable[[str], str], None)

    with pytest.raises(ValueError):
        slugify("!!!")
